/*****************************************************************************\
 *
 * Main-Windows.cpp
 *
 * Main-Windows - main function for Windows platform (SoWin)
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *          PCJohn (peciva _AT fit.vutbr.cz) - original terrain aplication from www.root.cz
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include <Inventor/SbTime.h>
#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/viewers/SoWinExaminerViewer.h>
#include <Inventor/Win/viewers/SoWinFlyViewer.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>

#include "perlin.h"

// Pro linker: Budeme linkovat s main,
// nikoliv WinMain, jak je na Windows zvykem,
// a aplikace bude vyuvat konzoli.
#pragma comment(linker, "/entry:\"mainCRTStartup\"")
#pragma comment(linker, "/subsystem:\"console\"")


//Objekt sceny
SoSeparator *root;

//Objekt kamery
SoCamera *kamera;

//Callback funkce, ktera zajisti dynamicke zmeny krajiny podle vzdalenosti od kamery
void LOD_Callback(void *data, SoSensor *sensor)
{
	float kameraX;
	float kameraY;
	float kameraZ;
	//Zjistime pozici kamery
	kamera->position.getValue().getValue(kameraX,kameraY,kameraZ);

	//Otestujeme krajinu a pokud doslo ke zmenam, prekreslime ji
	if (Terrain->test(kameraX,kameraZ,kameraY) == false)
		{
			root->removeAllChildren();

			// Nastav intenzitu globlnho svtla ve scn
			SoEnvironment *env = new SoEnvironment;
			env->ambientIntensity.setValue(0.5);
			root->addChild(env);

			// Svtlo
			SoDirectionalLight *sunLight = new SoDirectionalLight;
			sunLight->direction.setValue(0.35f, -1.0f, 0.25f);
			sunLight->intensity.setValue(1.0f);
			root->addChild(sunLight);

			//Vodni hladina
			// Nejprve materil, pak souadnice, a na zvr stripSet
			SoMaterial *waterMat = new SoMaterial;
			waterMat->ambientColor.setValue(0.f, 0.f, 1.f);
			waterMat->diffuseColor.setValue(0.f, 0.f, 0.f);
			root->addChild(waterMat);

			SoCoordinate3 *waterCoords = new SoCoordinate3;
			float waterLevel = (maxHeight - minHeight) / 5.f + minHeight;
			waterCoords->point.set1Value(0, SbVec3f(float(Terrain->x1), waterLevel, float(Terrain->y1)));
			waterCoords->point.set1Value(1, SbVec3f(float(Terrain->x2), waterLevel, float(Terrain->y1)));
			waterCoords->point.set1Value(2, SbVec3f(float(Terrain->x1), waterLevel, float(Terrain->y2)));
			waterCoords->point.set1Value(3, SbVec3f(float(Terrain->x2), waterLevel, float(Terrain->y2)));
			root->addChild(waterCoords);

			SoTriangleStripSet *waterStripSet = new SoTriangleStripSet;
			waterStripSet->numVertices.set1Value(0, 4); // nult prvek nastav na hodnotu 4,
					// to znamen, e budeme renderovat 2 trojhelnky (poet bod snme o
					// dva a zskme tak poet trojhelnk
			root->addChild(waterStripSet);

			//Vloz model krajiny
			Terrain->generate(root,kameraX,kameraZ,kameraY);
		}
}

int main(int argc, char **argv)
{
	// Zapne zobrazovani FPS
	putenv("COIN_SHOW_FPS_COUNTER=1");

  // Inicializuje Inventor
  HWND window = SoWin::init(argv[0]);
  if (window == NULL) exit(1);
	
	// Vytvori koren sceny
  root = createScene(argc, argv);
  root->ref();
	

	// Vytvori okno prohlizece
  SoWinExaminerViewer *viewer = new SoWinExaminerViewer(window);
  
  viewer->setSceneGraph(root);
  viewer->setHeadlight(headlight);
  viewer->setTitle(appTitle);
  viewer->show();

	//Ziska ukazatel na kameru (pro pouziti v callback funkci)
	kamera = viewer->getCamera();

	//Vytvori sensor a priradi mu callback funkci
	SbTime *cas = new SbTime();
	cas->setValue(0.1); //aktualizace krajiny 10x za sekundu
	SoTimerSensor *muj_sensor = new SoTimerSensor(LOD_Callback, NULL);
	muj_sensor->setInterval(cas->getValue());
	muj_sensor->schedule();

  // Zobrazi okna a rozjede renderovaci smycku
  SoWin::show(window);
  SoWin::mainLoop();
	
  // Uvolni pamet
  delete viewer;
  root->unref();

	return 0;
}