/*****************************************************************************\
 *
 * perlin.cpp
 *
 * perlin - generate terrain and create chunked-LOD structure
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *          PCJohn (peciva _AT fit.vutbr.cz) - original terrain aplication from www.root.cz
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include "perlin.h"


#ifdef _MSC_VER
float strtof(const char *nptr, char **endptr)  { return float(strtod(nptr, endptr)); }
#endif

const char *appTitle = "Perlin function terrain generator and chunked-LOD renderer";      // Titulek okna
bool headlight = false;                           // Defaultn svtlo (v pozici kamery)
SoSeparator* createScene(int argc, char **argv);  // Funkce vytvejc scnu

//Vzdalenost od ktere bude klesat detail segmentu
float LODdistance = 55;

//globalni promenne pro minimalni a maximalni vysku
float minHeight =  999999;
float maxHeight = -999999;

//pocet jiz vygenerovanych segmentu
int already_finished = 0;
int total_segments = 0;


//Nainportuju kameru
extern SoCamera *kamera;

// Globalni promenne a funkce,
// ktere jsou pouzity v modulu main-windows.cpp
TerrainPatch *Terrain;
tmpTerrainPatch *tmpTerrain;


/*!
  Funkce alokuje pamet pro vyskovou mapu.
 */
void tmpTerrainPatch::allocHmap(int dimension)
{
  if (this->hmap == NULL || this->dimension != dimension) {
    delete[] this->hmap;
    this->dimension = dimension;
  }
  this->hmap = new float[dimension*dimension];
}



/*!
  Funkce uvolni pamet, ktera byla alokovana pro vyskovou mapu.
 */

void tmpTerrainPatch::freeHmap()
{
  delete[] this->hmap;
  this->hmap = NULL;
}

/*!
  Funkce aplikuje na krajinu filter.
 */
void tmpTerrainPatch::filter(float q)
{
  int x, y;

  // dky zleva doprava
  for (y = 0; y < dimension; y++)
    for (x = 1; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // dky zprava doleva
  for (y = 0; y < dimension; y++)
    for (x = dimension-1; x >= 1; x--)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce zespoda nahoru
  for (y = 1; y < dimension; y++)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce shora dol
  for (y = dimension-1; y >= 1; y--)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;
}

/*!
  Funkce vraci index do vyskove mapy
 */
inline int tmpTerrainPatch::getHmapIndex(const int x, const int y)
{
  return y*dimension + x;
}

/*!
  Konstruktor
 */
tmpTerrainPatch::tmpTerrainPatch()
{
  hmap = NULL;
	x1 = -1;
	y1 = -1;
	x2 = -1;
	y2 = -1;
}

/*!
  Destruktor
 */
tmpTerrainPatch::~tmpTerrainPatch()
{
	reset();
}

/*!
  Funkce vraci hodnotu z vyskove mapy
 */
float tmpTerrainPatch::getHmapValue(const int x, const int y)
{
	if ((x >= 0) && (x < dimension) && (y >= 0) && (y < dimension))
		return hmap[y*dimension + x];
	else
		return 0;
}

/*!
  Funkce vygeneruje krajinu s defaulnimi hodnotami
 */
void tmpTerrainPatch::generateHeightMap()
{
  generateHeightMap(0,0,DEFAULT_TERRAIN_DIMENSION+1,DEFAULT_TERRAIN_DIMENSION+1,
		DEFAULT_TERRAIN_FILTER_Q);
}

/*!
  Funkce vygeneruje krajinu s vlastnimi hodnotami
 */
void tmpTerrainPatch::generateHeightMap(int _x1, int _y1, int _x2, int _y2, float filterQ)
{
	//Naplnime promenne rozmeru
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;


  // Alokujeme vkovou mapu
  allocHmap(x2-x1);

	// Vygenerujeme vyskovou mapu pomoci Perlinovy funkce
	int x,y;
	float pom;

  for (y=0; y<dimension; y++) {
	  for (x=0; x<dimension; x++) {
			pom = 150*PerlinNoise(float(x)/50,float(y)/50);
			if (pom > maxHeight) maxHeight = pom;
			if (pom < minHeight) minHeight = pom;
      hmap[getHmapIndex(x, y)] = pom;
	  }
		already_finished++;
		printf("\rHotovo: (%d%%)",already_finished*100/x2-x1);
  }

	//Vyfiltrovani krajiny
	filter(filterQ);

}

/*!
  Funkce znovunastavi krajinu.
 */
void tmpTerrainPatch::reset()
{
	freeHmap();
}


/*!
  Funkce vraci index do vyskove mapy hmap pro souradnice [x,y].
*/
inline int TerrainPatch::getHmapIndex(const int x, const int y)
{
  return y*dimension + x;
}

/*!
  Funkce otestuje, zda je scena spravne zobrazena.
*/
bool TerrainPatch::test(float cameraX, float cameraY, float cameraZ)
{
	bool tmp_output;

	double midx = double(x1) + (x2 - x1) / 2;
	double midy = double(y1) + (y2 - y1) / 2;
	double midz = hmap[getHmapIndex(dimension/2,dimension/2)];
	double deltax = midx - cameraX;
	double deltay = midy - cameraY;
	double deltaz = midz - cameraZ;
	double distance2D = sqrt(deltax*deltax+deltay*deltay);
	double distance = sqrt(distance2D*distance2D + deltaz*deltaz);

	if ((((chunk_error * LODdistance) / distance) <= 1.0) || ((x2-x1) <= dimension))
		{ //chyba neni velka nebo jsme na konci stromu => tento segment ma byt vykreslen
			if (model_showed == false)
			{
				model_showed = true;
				return false;
			}
			else
			{
				return true;
			}
		}
	else
		{ //pokracujeme do dalsich podsegmentu
			tmp_output = this->subs0->test(cameraX,cameraY,cameraZ);
			tmp_output = this->subs1->test(cameraX,cameraY,cameraZ) && tmp_output;
			tmp_output = this->subs2->test(cameraX,cameraY,cameraZ) && tmp_output;
			tmp_output = this->subs3->test(cameraX,cameraY,cameraZ) && tmp_output;
			if (model_showed == true)
			{
				model_showed = false;
				return false;
			}
			else
			{
				return true && tmp_output;
			}
		}
}

/*!
  Funkce prida ctverec krajiny do celkove krajiny.
*/
void TerrainPatch::generate(SoSeparator *root, float cameraX, float cameraY, float cameraZ)
{
	if (model_showed == true)
	{
		addModel(root);
	}
	else
	{
		subs0->generate(root,cameraX,cameraY,cameraZ);
		subs1->generate(root,cameraX,cameraY,cameraZ);
		subs2->generate(root,cameraX,cameraY,cameraZ);
		subs3->generate(root,cameraX,cameraY,cameraZ);
	}
}



/*!
	Tato funkce naplni vyskovou mapu aktualniho segmentu. Data prebira z pomocneho segmentu.
*/
void TerrainPatch::generateHeightMap(int _x1, int _y1, int _x2, int _y2, int seg_dimension, int hq)
{
	//Naplnime promenne rozmeru
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;


  // Alokujeme vkovou mapu
  allocHmap(seg_dimension);

	// Vygenerujeme vyskovou mapu z pomocneho segmentu
	int x,y;
	chunk_error = float(x2 - x1) / (dimension - 1);
	int step = (x2 - x1) / (dimension - 1);
  for (y=0; y<dimension; y++) {
	  for (x=0; x<dimension; x++) {
			hmap[getHmapIndex(x, y)] = tmpTerrain->getHmapValue(x1 + x*step,y1 + y*step);
	  }
  }

	// Vygenerujeme poduzly
	if ((x2 - x1) <= dimension)
	{ //uz jsme na nejnizsi urovni

	}
	else
	{ //musime vygenerovat poduzly
		allocSubs();
		int midx = (x2 - x1) / 2;
		int midy = (y2 - y1) / 2;
		subs0->generateHeightMap(x1,y1,x1+midx,y1+midy,seg_dimension,hq);
		subs1->generateHeightMap(x1+midx,y1,x2,y1+midy,seg_dimension,hq);
		subs2->generateHeightMap(x1,y1+midy,x1+midx,y2,seg_dimension,hq);
		subs3->generateHeightMap(x1+midx,y1+midy,x2,y2,seg_dimension,hq);
	}


	//Vygenerovani modelu
	generateModel(float(x2-x1),hq);

	//Ukazatel hotovych segmentu
	already_finished++;
	printf("\rHotovo: %d (%d%%)",already_finished,already_finished*100/total_segments);
}



/*!
	Tato funkce naplni vyskovou mapu aktualniho segmentu s pouzitim defaultnich hodnot. Data prebira z pomocneho segmentu.
*/
void TerrainPatch::generateHeightMap()
{
  generateHeightMap(0,0,DEFAULT_TERRAIN_DIMENSION,DEFAULT_TERRAIN_DIMENSION,
		DEFAULT_SEGMENT_DIMENSION,DEFAULT_TERRAIN_HQ);
}



/*!
  Funkce vytvori zobrazitelny graf sceny modelu krajiny
  zalozene na vyskove mape. Parametr size udava velikost krajiny 
	(v prostorovych souradnicich)
  a parametr hq udava metodu vytvoreni modelu krajiny.
  0 znaci standardni tesselaci, tj. dva trojuhelniky na ctverec krajiny,
  1 je podobna predchozi volbe, ale zlom mezi temito trojuhelnky je
  vytvaren podle toho, ve kterem smeru je v dane dlazdici vetsi zmena vysky,
  2 dlazdice je rendrovana ctyrmi trojuhelniky.
*/
void TerrainPatch::generateModel(float size, int hq)
{
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	model->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  model->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  model->addChild(mbind);

  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (hq != 2) ? SbSqr(dimension) : SbSqr(dimension)+SbSqr(dimension-1);
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();


  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi,yi;
  float sx = size / (dimension-1);
  float sy = sx;
  SbVec3f midp;

  if (hq != 2) {
    for (yi=0; yi<dimension; yi++) {
      for (xi=0; xi<dimension; xi++) {

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,yi)], float(y1) + yi*sy);
        *(n++) = calculateNormal(xi, yi, sx, sy);

      }
    }
  } else {
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {

        // Vpoet prostednho bodu tverce
        midp.setValue(float(x1) + (xi+0.5f) * sx,
                      (hmap[getHmapIndex(xi,yi)] + hmap[getHmapIndex(xi+1,yi)] +
                      hmap[getHmapIndex(xi,yi+1)]+ hmap[getHmapIndex(xi+1,yi+1)]) / 4.f,
                      float(y1) + (yi+0.5f) * sy);

        // Spotme si normly pro tyi rohov body i pro
        // stedov bod
        SbVec3f n1 = calculateNormal(xi, yi, sx, sy);
        SbVec3f n2 = calculateNormal(xi, yi+1, sx, sy);
        SbVec3f n3 = calculateNormal(xi+1, yi+1, sx, sy);
        SbVec3f n4 = calculateNormal(xi+1, yi, sx, sy);
        SbVec3f nmid = n1+n2+n3+n4;
        nmid.normalize();

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,yi)], float(y1) + yi*sy);
        *(n++) = n1;

        // souadnice a normla pro prostedn bod ve tverci krajiny
        *(c++) = midp;
        *(n++) = nmid;
      }
      // posledn bod v dku
      *(c++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,yi)], float(y1) + yi*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }

    // Posledn ada vertex
    for (xi=0; xi<dimension; xi++) {
      *(c++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,yi)], float(y1) + yi*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }
  }


  // Nakonfiguruj triangleStripSet
  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  switch (hq) {
  case 0: num = dimension * (dimension-1) * 2 + dimension-1; break;
  case 1: num = dimension * (dimension-1) * 5; break;
  case 2: num = (dimension-1) * (dimension-1) * 10; break;
  }
  stripSet->coordIndex.setNum(num);
  stripSet->materialIndex.setNum(num);

  // Prvnmu ze strip vdy pia 5 souadnic (vytvo 3 trojhelnky)
  // a druhmu 3 (vytvo jeden trojhelnk).
  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();
  switch (hq) {
  case 0: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension; xi++) {
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
      }
      *(ci++) = -1;     // konec triangle stripu
      *(mi++) = -1;
    }
    break;
  }
  case 1: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Potej zmnu v jednotlivch smrech dladice,
        // abychom zjistili, ve kterm smru bude dladice rozdlena.
        bool orderChange = fabsf(c[i2+1][1]-c[i1][1]) > fabsf(c[i2][1]-c[i1+1][1]);
        if (orderChange) {
          int tmp = i2;
          i2 = i1;
          i1 = tmp;
        }

        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
        *(ci++) = i2;   // prav horn bod tverce
        *(ci++) = i1;   // prav spodn bod tverce
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // prav horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // prav spodn bod tverce
        *(ci++) = -1;
        *(mi++) = -1;
      }
      i1++;
      i2++;
    }
    break;
  }
  case 2: {
    c = ctemp;
    int tmp1,tmp2,tmp3;
    int i1 = 0;
    int i2 = 2*dimension-1;
    for (yi=0; yi<dimension-2; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Generovn souadnicovch index
        // (vdy dva triangleStripSety, prvn sloen ze t,
        // druh z jednoho trojhelnku)
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i1+2; // prav spodn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;
        *(ci++) = i2;   // lev horn bod
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;

        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight); // prostedn bod
        *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight); // prav spodn bod
        *(mi++) = tmp3 = getMaterialIndex(c[i2+2][1], minHeight, maxHeight); // prav horn bod
        *(mi++) = -1;
        *(mi++) = tmp1; // lev horn bod
        *(mi++) = tmp2; // prostedn bod
        *(mi++) = tmp3; // prav horn bod
        *(mi++) = -1;

        i1 += 2;
        i2 += 2;
      }
      i1++;
      i2++;
    }

    // Posledn dek
    for (xi=0; xi<dimension-1; xi++) {
      *(ci++) = i2;
      *(ci++) = i1;
      *(ci++) = i1+1;
      *(ci++) = i1+2;
      *(ci++) = i2+1; // zde zmna oproti pedchoz smyce
      *(ci++) = -1;
      *(ci++) = i2;
      *(ci++) = i1+1;
      *(ci++) = i2+1; // zde zmna
      *(ci++) = -1;

      *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);
      *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight);
      *(mi++) = tmp3 = getMaterialIndex(c[i2+1][1], minHeight, maxHeight); // zde zmna
      *(mi++) = -1;
      *(mi++) = tmp1;
      *(mi++) = tmp2;
      *(mi++) = tmp3;
      *(mi++) = -1;

      i1 += 2;
      i2 += 1;
    }
    break;
  }}
  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  model->addChild(normals);
  model->addChild(coords);
  model->addChild(stripSet);

	//----------------------------------------------------------
	//Dogeneruji svisle hranice, aby nevznikaly "diry" v krajine
	//----------------------------------------------------------

	//Definuji materialy
	SoMaterial *material1 = new SoMaterial;
	material1->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material1->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material1->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material1->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material1->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material1->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material1->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material1->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	border->addChild(material1);


	//Zpusob ulozeni normal
  SoNormalBinding *nbind1 = new SoNormalBinding;
  nbind1->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  border->addChild(nbind1);

	//Zpusob ulozeni materialu
	SoMaterialBinding *mbind1 = new SoMaterialBinding;
	mbind1->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
	border->addChild(mbind1);

	SoCoordinate3 *coords1 = new SoCoordinate3;
	SoNormal *normals1 = new SoNormal;
	int num1 = (8*dimension-4);
	coords1->point.setNum(num1);
	normals1->vector.setNum(num1);
	SbVec3f *c1 = coords1->point.startEditing();
	SbVec3f *ctemp1 = c1;
	SbVec3f *n1 = normals1->vector.startEditing();

	sx = size / (dimension-1);
	sy = sx;

	//Generuji trojuhelniky pro jednotlive strany segmentu do jedineho stripSetu
	for (xi=0; xi<dimension-1; xi++) 
	{
		*(c1++) = SbVec3f(float(x1), hmap[getHmapIndex(0,xi)], float(y1) + xi*sy);
		*(n1++) = calculateNormal(0, xi, sx, sy);

		*(c1++) = SbVec3f(float(x1), minHeight, float(y1) + xi*sy);
		*(n1++) = calculateNormal(0, xi, sx, sy);
	}
	for (xi=0; xi<dimension-1; xi++) 
	{
		*(c1++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,dimension-1)], float(y2));
		*(n1++) = calculateNormal(xi, dimension-1, sx, sy);

		*(c1++) = SbVec3f(float(x1) + xi*sx, minHeight, float(y2));
		*(n1++) = calculateNormal(xi, dimension-1, sx, sy);
	}

	for (xi=dimension - 1; xi >= 0; xi--) 
	{
		*(c1++) = SbVec3f(float(x2), hmap[getHmapIndex(dimension-1,xi)], float(y1) + xi*sy);
		*(n1++) = calculateNormal(dimension-1, xi, sx, sy);

		*(c1++) = SbVec3f(float(x2), minHeight, float(y1) + xi*sy);
		*(n1++) = calculateNormal(dimension-1, xi, sx, sy);
	}

		
	for (xi=dimension - 1; xi >= 0; xi--) 
	{
		*(c1++) = SbVec3f(float(x1) + xi*sx, hmap[getHmapIndex(xi,0)], float(y1));
		*(n1++) = calculateNormal(xi, 0, sx, sy);

		*(c1++) = SbVec3f(float(x1) + xi*sx, minHeight, float(y1));
		*(n1++) = calculateNormal(xi, 0, sx, sy);
	}

	//Vytvorim StripSet a naplnim ho
	SoIndexedTriangleStripSet *triangleSet = new SoIndexedTriangleStripSet;
	triangleSet->coordIndex.setNum(num1+1);
	triangleSet->materialIndex.setNum(num1+1);

	int32_t *ci1 = triangleSet->coordIndex.startEditing();
	int32_t *mi1 = triangleSet->materialIndex.startEditing();
	c1 = ctemp1;
	for (int i1=0; i1<num1; i1++) 
	{
		*(ci1++) = i1;
		*(mi1++) = getMaterialIndex(c1[i1][1], minHeight, maxHeight);
	}
	*(ci1++) = -1; //Konec stripSetu
	*(mi1++) = -1;

	triangleSet->coordIndex.finishEditing();
	triangleSet->materialIndex.finishEditing();
	coords1->point.finishEditing();
	normals1->vector.finishEditing();
	
	//Vlozim stripSet do modelu hranice
	border->addChild(normals1);
	border->addChild(coords1);
	border->addChild(triangleSet);
}


/*!
	Tato funkce prida model ho stromu root.
*/
void TerrainPatch::addModel(SoSeparator *root)
{
	root->addChild(model);
	root->addChild(border);
}

/*!
  Funkce alokuje model.
 */
void TerrainPatch::allocModel()
{
  if (this->model != NULL) {
		freeModel();
	}
	else
	{
    this->model = new SoSeparator;
    this->border = new SoSeparator;
	}
	model->ref();
	border->ref();
}



/*!
  Funkce uvoln pamt, kter byla alokovna pro model.
 */
void TerrainPatch::freeModel()
{
  this->model->unref();
  this->border->unref();
}

/*!
  Funkce alokuje novou vkovou mapu, kter je vdy
  tvercov. Jej rozmr je zadn v parametru dimension,
  kter mus bt mocninou 2 plus 1.
 */
void TerrainPatch::allocHmap(int dimension)
{
  if (this->hmap == NULL || this->dimension != dimension) {
    delete[] this->hmap;
    this->dimension = dimension;
  }
  this->hmap = new float[dimension*dimension];
}



/*!
  Funkce uvoln pam, kter byla alokovna pro vkovou mapu.
 */
void TerrainPatch::freeHmap()
{
  delete[] this->hmap;
  this->hmap = NULL;
}

/*!
  Funkce alokuje poduzly
 */
void TerrainPatch::allocSubs()
{
  delete subs0;
  delete subs1;
  delete subs2;
  delete subs3;
	subs0 = new TerrainPatch();
	subs1 = new TerrainPatch();
	subs2 = new TerrainPatch();
	subs3 = new TerrainPatch();
}

/*!
  Funkce uvolni poduzly.
 */
void TerrainPatch::freeSubs()
{
  delete subs0;
  delete subs1;
  delete subs2;
  delete subs3;
}


/*!
  Funkce vypot normlu v danm bod vkov mapy.
 */
SbVec3f TerrainPatch::calculateNormal(const int x, const int y, const float sx, const float sy)
{
  float dx,dy;
	int step = (x2 - x1) / (dimension - 1);
  // hel v ose X
  if (x>=1 && x<=dimension-2)
    // Vpoet ze dvou sousednch bod
    dx = (hmap[getHmapIndex(x-1,y)] - hmap[getHmapIndex(x+1,y)]) / 2.f;
  else
    // Vpoet na okraji mapy (pouziji se hodnoty z pomocneho segmentu pro lepsi navazovani segmentu)
    if (x<1)  dx = (tmpTerrain->getHmapValue((x-1)*step + x1,y*step + y1) - hmap[getHmapIndex(x+1,y)]) / 2;
    else  dx = (hmap[getHmapIndex(x-1,y)] - tmpTerrain->getHmapValue((x+1)*step + x1,y*step + y1)) / 2;

  // hel v ose Y
  if (y>=1 && y<=dimension-2)
    dy = (hmap[getHmapIndex(x,y-1)] - hmap[getHmapIndex(x,y+1)]) / 2.f;
  else
    // Vpoet na okraji mapy (pouziji se hodnoty z pomocneho segmentu pro lepsi navazovani segmentu)
    if (y<1)  dy = (tmpTerrain->getHmapValue(x*step + x1,(y-1)*step + y1) - hmap[getHmapIndex(x,y+1)]) / 2;
    else  dy = (hmap[getHmapIndex(x,y-1)] - tmpTerrain->getHmapValue(x*step + x1,(y+1)*step + y1)) / 2;

/*
	// hel v ose X
  if (x>=1 && x<=dimension-2)
    // Vpoet ze dvou sousednch bod
    dx = (hmap[getHmapIndex(x-1,y)] - hmap[getHmapIndex(x+1,y)]) / 2.f;
  else
    // Vpoet na okraji mapy
    if (x<1)  dx = hmap[getHmapIndex(x,y)] - hmap[getHmapIndex(x+1,y)];
    else  dx = hmap[getHmapIndex(x-1,y)] - hmap[getHmapIndex(x,y)];
  // hel v ose Y
  if (y>=1 && y<=dimension-2)
    dy = (hmap[getHmapIndex(x,y-1)] - hmap[getHmapIndex(x,y+1)]) / 2.f;
  else
    if (y<1)  dy = hmap[getHmapIndex(x,y)] - hmap[getHmapIndex(x,y+1)];
    else  dy = hmap[getHmapIndex(x,y-1)] - hmap[getHmapIndex(x,y)];
*/

  SbVec3f r(dx/sx, 1.f, dy/sy);
  r.normalize();
  return r;
}



/*!
  Funkce vrt index materilu pro dan bod vkov mapy.
 */
int TerrainPatch::getMaterialIndex(const float height, const float minHeight, const float maxHeight)
{
  float part = (maxHeight-minHeight) / 5.f;
  if (height <= minHeight+3.f*part)  return 1; // zelen materil
  if (height <= minHeight+4.f*part)  return 2; // utrovit materil
  return 3; // snh
}



/*!
  Konstruktor.
*/
TerrainPatch::TerrainPatch()
{
  hmap = NULL;
	subs0 = NULL;
	subs1 = NULL;
	subs2 = NULL;
	subs3 = NULL;
	model = NULL;
	border = NULL;
	x1 = -1;
	y1 = -1;
	x2 = -1;
	y2 = -1;
	model_showed = false;
	chunk_error = 0;
	allocModel();
}



/*!
  Destruktor.
*/
TerrainPatch::~TerrainPatch()
{
  reset();
}



/*!
  Inicializuje objekt do podobnho stavu jako pi vytvoen
  objektu, tzn. minimln vyuit zdroj.
  Funkce se pouv pedevm pro uvolnn pamti
  vkov mapy, nen-li ji dle potebn.
*/
void TerrainPatch::reset()
{
  freeHmap();
	freeSubs();
	freeModel();
}



// Funkce pouit v sablonov funkci getParam pro konverzi
// string do rznch typ.
void strtoX(int &r, const char *nptr, char **endptr)
{
  r = strtol(nptr, endptr, 10);
}
void strtoX(float &r, const char *nptr, char **endptr)
{
  r = strtof(nptr, endptr);
}
void strtoX(bool &r, const char *nptr, char **endptr)
{
  int i;
  strtoX(i, nptr, endptr);
  r = (i != 0);
}

// Sablonov funkce pro ten parametr rznch typ z pkazov dky
template<class Type>
bool getParam(int argc, char **argv, int &position,
              const char *param, Type &result)
{
  if (argc > position && strcmp(argv[position], param) == 0) {
    position++;
    if (argc > position) {
      char *errp;
      Type tmp;
      strtoX(tmp, argv[position], &errp);
      if (*(argv[position]) != '\0' && *errp == '\0') {
        result = tmp;
        position++;
        return true;
      } else {
#if !defined(__WIN32__) && !defined(_WIN32)
        printf("Chybn argument \"%s\".", param);
#else
        printf("Chybny argument \"%s\".", param);
#endif
        exit(EXIT_FAILURE);
      }
    } else {
#if !defined(__WIN32__) && !defined(_WIN32)
      printf("Chybjc hodnota pro parametr \"%s\".", param);
#else
      printf("Chybejici hodnota pro parametr \"%s\".", param);
#endif
      exit(EXIT_FAILURE);
    }
  }
  return false;
}


// Funkce vytvo graf scny pro nasi aplikaci.
SoSeparator* createScene(int argc, char **argv)
{
  int i = 1;

  // Vpis npovdy
  if (argc > 1 && (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
    printf("Pouziti: %s [-h|--help] [-seed int] [-dimension int] [-size float]\n"
           "         [-heightQ float] [-filterQ float] [-hq int]\n"
           "-h, --help        tato obrazovka\n"
           "-seed int         nastaveni generatoru nahodnych cisel\n"
           "-size int         velikost krajiny\n"
           "-dimension int    velikost segmentu\n"
           "-distance float   vzdalenost od ktere se snizuje LOD\n"
           "-filterQ float    parametr filtrovaci funkce\n"
           "-hq int           hodnota 0,1 a 2 meni rezim tesselace krajiny\n"
           "                  (0 standard, 1 adaptivni smer hrany, 2 ctyri trojuhelniky)\n"
           "Priklad pouziti (se standardnimi hodnotami parametru krome seed):\n"
           "%s -seed 10 -size 512 -dimension 33 -distance 100.0 -filterQ 0.5 -hq 0\n",
           argv[0], argv[0]);
    exit(EXIT_SUCCESS);
  }

  // Inicializace genertoru nhodnch sel
  int seed;
  if (getParam<int>(argc, argv, i, "-seed", seed))
    srand(seed);
  else
    srand((unsigned int)time(NULL));

	//Inicializuje Perlinovu funkci (konstanty)
	PerlinInit();

  // Vytvo koen scny
  SoSeparator *root = new SoSeparator;
  root->ref();

  // Nastav intenzitu globlnho svtla ve scn
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.5);
  root->addChild(env);

  // Svtlo
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(0.35f, -1.0f, 0.25f);
  sunLight->intensity.setValue(1.0f);
  root->addChild(sunLight);

  // Model krajiny
  int size = DEFAULT_TERRAIN_DIMENSION;
  int dimension = DEFAULT_SEGMENT_DIMENSION;
  float filterQ = DEFAULT_TERRAIN_FILTER_Q;
  int hq = DEFAULT_TERRAIN_HQ;
	LODdistance = float(sqrt(3)*dimension);
  getParam<int>(argc, argv, i, "-size", size);
  getParam<int>(argc, argv, i, "-dimension", dimension);
  getParam<float>(argc, argv, i, "-distance", LODdistance);
  getParam<float>(argc, argv, i, "-filterQ", filterQ);
  getParam<int>(argc, argv, i, "-hq", hq);

  // Zkontroluj, zda byly zparsovny spn vechny argumenty z command-line.
  if (i != argc) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Chybn parametr %s.\n", argv[i]);
#else
    printf("Chybny parametr %s.\n", argv[i]);
#endif
    exit(EXIT_FAILURE);
  }

  // Zkontroluj platnost parametr danch uivatelem.
  // size mus bt mocnina 2
  if (int(pow(2, int(log(size)/log(2.f)+0.5f))+0.5f) != size) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr size mus bt mocnina dvou.\n"
           "Napklad 256,512,1024,2048 a podobn.\n");
#else
    printf("Parametr size musi byt mocnina dvou.\n"
           "Napriklad 256,512,1024,2048 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // Zkontroluj platnost parametr danch uivatelem.
  // dimension mus bt mocnina 2 plus 1
  if (int(pow(2, int(log(dimension-1)/log(2.f)+0.5f))+0.5f) != dimension-1) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr dimension mus bt mocnina dvou plus 1.\n"
           "Napklad 17,33,65,129 a podobn.\n");
#else
    printf("Parametr dimension musi byt mocnina dvou plus 1.\n"
           "Napriklad 17,33,65,129 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // filterQ mus bt v rozsahu od 0.f do 1.f
  if (filterQ < 0.f || filterQ > 1.f) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr filterQ mus bt v rozsahu 0.0 a 1.0.\n"
           "Pklad: 0.5, 0.36 a podobn.\n");
#else
    printf("Parametr filterQ musi byt v rozsahu 0.0 az 1.0.\n"
           "Priklad: 0.5, 0.36 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // Omez hq na hodnoty 0,1 a 2
  if (hq < 0 || hq > 2)  hq = 2;

  // Generovn krajiny
  printf("Generuji krajinu Perlinovou funkci:\n"
         "velikost terenu:    %i\n"
         "velikost segmentu:  %i\n"
         "LOD vzdalenost:     %.2f\n"
         "filterQ:            %.2f\n"
         "hq:                 %i\n",
         size,dimension, LODdistance, filterQ, hq); 
	
	//Vygenerovani krajiny do pomocneho segmentu
	tmpTerrain = new tmpTerrainPatch();
	already_finished = 0;
	printf("Generovani vyskove mapy:\n");
	tmpTerrain->generateHeightMap(0,0,size+1,size+1,filterQ);
	Terrain = new TerrainPatch();
	for (int ii = size/(dimension-1);ii >= 1;ii=ii/2)
	{
		total_segments += ((ii) * (ii));
	}

	//Vygenerovani Quad-Tree
	already_finished = 0;
	printf("\nGenerovani segmentu krajiny:\n");
	Terrain->generateHeightMap(0,0,size,size,dimension,hq);

	//Prvotni vygenerovani modelu
	Terrain->test(float(size) / 2,float(size),0);
	Terrain->generate(root,float(size) / 2,float(size),0);

	//Pomocny segment uz nebudeme potrebovat
	tmpTerrain->~tmpTerrainPatch();
  root->unrefNoDelete();
  return root;
}