/*****************************************************************************\
 *
 * perlin.h
 *
 * perlin - header for perlin.cpp (generate terrain and create chunked-LOD structure)
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *          PCJohn (peciva _AT fit.vutbr.cz) - original terrain aplication from www.root.cz
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#ifndef __PERLIN_H
#define __PERLIN_H

#include "perlin_f.h"

#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

extern const char *appTitle; // Titulek okna
extern bool headlight; // Defaultn svtlo (v pozici kamery)
extern SoSeparator* createScene(int argc, char **argv);  // Funkce vytvejc scnu


//Velikost cele krajiny
#define DEFAULT_TERRAIN_DIMENSION 512

// Velikost vyskove mapy segmentu zvetsen o jednu
// (krajina 32x32 se konstruuje ze 33x33 bod)
#define DEFAULT_SEGMENT_DIMENSION  33

// Filtrovaci koeficient pro finalni vyhlazeni krajiny.
// Vyssi cislo odpovida vyssi "hrbolatosti",
// nizsi - vice se to vyfiltruje.
#define DEFAULT_TERRAIN_FILTER_Q  0.5f

// HQ (high quality) uruje, jakm zpsobem bude rendrovna vsledn krajina.
// HQ=0 znamen, e kad tverec bude rendrovn jako dva trojhelnky,
// HQ=1 je podobn pedchoz volb, ale zlom mezi tmito trojhelnky je
// vytven podle toho, ve kterm smru je v dan dladici vt zmna vky,
// HQ=2 pid do stedu tverce dal vertex a vyrendruje tverec tymi
// trojhelnky. Krajina bude mt dvojnsobn mnostv trojhelnk,
// ale stane se hlad.
#define DEFAULT_TERRAIN_HQ  0

//Vzdalenost od ktere bude klesat detail segmentu
extern float LODdistance;

//globalni promenne pro minimalni a maximalni vysku
extern float minHeight;
extern float maxHeight;

//pocet jiz vygenerovanych segmentu
extern int already_finished;
extern int total_segments;


/*!
  Tda pedstavujc pomocn tverec v krajin.
*/
class tmpTerrainPatch {
private:
  void allocHmap(int dimension);
  void freeHmap();
  void filter(float q);

protected:
  float *hmap;
  int dimension;
	int x1,y1,x2,y2;
  inline int getHmapIndex(const int x, const int y);
public:
  tmpTerrainPatch();
  ~tmpTerrainPatch();

	float getHmapValue(const int x, const int y);
  void generateHeightMap();
  void generateHeightMap(int _x1, int _y1, int _x2, int _y2, float filterQ);

  void reset();
};

/*!
  Tda pedstavujc segment v krajin.
*/
class TerrainPatch {
private:
  void allocModel();
  void freeModel();
  void allocHmap(int dimension);
  void freeHmap();
  void allocSubs();
  void freeSubs();
  SbVec3f calculateNormal(const int x, const int y, const float sx, const float sy);

protected:
  float *hmap;
	SoSeparator *model;
	SoSeparator *border;
	bool model_showed;
	TerrainPatch *subs0;
	TerrainPatch *subs1;
	TerrainPatch *subs2;
	TerrainPatch *subs3;
  int dimension;
	float chunk_error;

  inline int getHmapIndex(const int x, const int y);
  int getMaterialIndex(const float height, const float minHeight, const float maxHeight);

public:
	int x1,y1,x2,y2;
  TerrainPatch();
  ~TerrainPatch();

  void generateHeightMap();
  void generateHeightMap(int _x1, int _y1, int _x2, int _y2, int seg_dimension, int hq);

  void generateModel(float size, int hq);

	void generate(SoSeparator *root, float cameraX, float cameraY, float cameraZ);
	bool test(float cameraX, float cameraY, float cameraZ);
  void reset();
	void addModel(SoSeparator *root);
};

extern TerrainPatch *Terrain;

#endif